/**\
|**| =====================================================================
|**|
|**|	Types Layout.c
|**|
|**|	This file contains the calls that this sample needs to make
|**|	the QuickDraw GX shell work correctly.
|**|
|**|	QuickDraw GX Libraries Used:
|**|	"ColorLibrary.c", "FontLibrary.c", "GraphicsDebugLibrary.c",
|**|	"LayoutLibrary.c", "ShapeLibrary.c", and "TransformLibrary.c".
|**|
|**|	6/96 bob	Updated #includes to support changed GX Library names.
|**|				Updated the copyright date.
|**|
|**|	1990 - 1996  Apple Computer, Inc.
|**|	All rights reserved.
|**|
|**| =====================================================================
\**/


#include "QDGX shell.h"
#include <GXLayout.h>
#include "LayoutLibrary.h"


/**\
|**| ---------------------------------------------------------------------
|**| PROTOTYPES
|**| ---------------------------------------------------------------------
\**/

// funtions required by shell

void	DoSetup			(void);
void	DoDraw			(WindowPtr wind, Boolean updating);
OSErr	DoCreateNew		(void);
void	DoDispose		(WindowPtr wind);
void	DoIdle			(WindowPtr wind);
void	DoTeardown		(void);
void	DoClick			(WindowPtr wind, Point p);

// private functions

OSErr	DoWindowInit		(WindowPtr wind);
void	CreateSampleImage	(WindowPtr wind);


/**\
|**| ---------------------------------------------------------------------
|**| ENUMS
|**| ---------------------------------------------------------------------
\**/
enum { rWindResource = 128 };


/**\
|**| ---------------------------------------------------------------------
|**| GLOBALS
|**| ---------------------------------------------------------------------
\**/
// If gDebugging = TRUE, graphics library errors and notices will be posted.  This
// functionality will only work with the "debugging" version of QuickDraw GX.
// If the debugging version is not installed, nothing bad will happen, but these
// functions will not work. 

Boolean		gDebugging = true;

// Set  "gGiveMeValidation" to TRUE if you want receive run-time validation.

Boolean		gGiveMeValidation = true;


// gGraphicsHeapSize sets the size of the graphics heap created by calling the
// GXNewGraphicsClient routine in main () within QuickDraw GX shell.c.  You can determine
// the amount of graphics heap required by using GraphicsBug.  I'm giving it 600K,
// since we need abunch if we have several windows open.

long		gGraphicsHeapSize = 600;

// gOurPrintingOverrideUPP is a universal proc pointer for our printing event
// override.  This is so that our override can be native PowerPC code if necessary.

GXPrintingEventUPP	gOurPrintingOverrideUPP;



/**\
|**| ---------------------------------------------------------------------
|**| DoSetup()
|**| Here's where we initialize any global variables our application needs.
|**| We have only one at this time -- the universal proc pointer for
|**| our printing override.
|**| ---------------------------------------------------------------------
\**/
void DoSetup (void)
{	// Initialize our printing event override UPP
	gOurPrintingOverrideUPP = NewGXPrintingEventProc(MyPrintingEventOverride);
}


/**\
|**| ---------------------------------------------------------------------
|**| DoDraw()
|**| Draw the contents of the window.  The first parameter is the window
|**| to draw, and the second parameter is true if we're updating an existing
|**| image.  If that's the case, we don't want to change anything, but
|**| just draw what's already there.
|**| ---------------------------------------------------------------------
\**/
void DoDraw (WindowPtr wind, Boolean updating)
{
	 #pragma unused (updating)
	 GXDrawShape (GetDocShape(wind));
}


/**\
|**| ---------------------------------------------------------------------
|**| DoCreateNew()
|**| This routine is called when a window needs to be created.
|**| ---------------------------------------------------------------------
\**/
OSErr DoCreateNew (void)
{
	OSErr		err = noErr;
	WindowPtr	wind;
	
// Get and create our window from the resource fork

	wind = GetNewWindow(rWindResource, nil, (WindowPtr)-1L);

// Attach a default gxViewPort to it, create and iInitialize our
// private data for it, and add a sample image to its page shape.

	if ( wind == NULL )
		return (MemError());

	GXIgnoreGraphicsNotice(transform_already_set);
	SetDefaultViewPort(GXNewWindowViewPort(wind));			
	GXPopGraphicsNotice();
	
	err = DoWindowInit(wind);
	if ( err != noErr )
		return err;
	
	CreateSampleImage(wind);
	return err;
}


/**\
|**| ---------------------------------------------------------------------
|**| DoDispose()
|**| This routine is called when a window needs to be disposed of.
|**| ---------------------------------------------------------------------
\**/
void DoDispose (WindowPtr wind)
{
	TH_Doc	doc;
	
// You should always dispose of your GX graphics objects before tossing your window.
// Why?  It's generally good form and this approach guarantees that everything is
// disposed.  If you had not disposed of everything, the call to DisposeWindow should
// dispose of the objects. If you are running the debugging version of QuickDraw GX
// with notices set, you will receive a notice that you had not disposed of everything.
// You can turn notices on in this file by setting gDebugging = TRUE (above).
	
	if ( wind != NULL )
	{
		doc = (TH_Doc)GetWRefCon(wind);		// Remember, this is where we stored our private data.
		GXDisposeShape(GetDocShape(wind)); 	// Dispose of this doc's shape.
		GXDisposeJob(GetDocJob(wind));		// Dispose of this doc's print job.
		DisposHandle((Handle) doc);			// Dispose of our private data.
		DisposeWindow(wind);				// Dispose of the window.
	}
}


/**\
|**| ---------------------------------------------------------------------
|**| DoIdle()
|**| This routine is called to do things while idling through the event loop.
|**| ---------------------------------------------------------------------
\**/
void DoIdle (WindowPtr wind)
{
}


/**\
|**| ---------------------------------------------------------------------
|**| DoTeardown()
|**| This routine is called just before we quit to remove anything 
|**| persistent that might have been setup by DoSetup().
|**| ---------------------------------------------------------------------
\**/
void DoTeardown (void)
{
	DisposeRoutineDescriptor(gOurPrintingOverrideUPP);
}


/**\
|**| ---------------------------------------------------------------------
|**| DoClick()
|**| ---------------------------------------------------------------------
\**/
void DoClick(WindowPtr window, Point p)
{
}





/**\
|**| ---------------------------------------------------------------------
|**| DoWindowInit()
|**| In this function we create and initialize the the private document
|**| structure for a new window.  This structure contains the print job and
|**| the shape which is drawn in the window.  We store this data in a handle
|**| and hang it off the window's refCon field for easy retrieval.  By doing
|**| this, rather than using globals, we can create many windows containing
|**| unique print jobs and shapes.
|**| ---------------------------------------------------------------------
\**/
OSErr DoWindowInit (WindowPtr wind)
{
	OSErr	err = noErr;
	gxJob	docJob;
	gxShape	docPage;
	TH_Doc	windDoc;


// Create the page shape. We set the unique items attribute to make sure that each item
// added to the picture has a unique reference. If this attribute was not set, we would
// not see all copies of anything we add to the shape multiple times -- we'd just see
// the last version added.		

	docPage = GXNewShape(gxPictureType);
	GXSetShapeAttributes(docPage, (GXGetShapeAttributes(docPage) | gxUniqueItemsShape));
	
	
// Create a print job for this document.  This will be the same as the system default until
// the user goes through the dialogs for Page Setup or Print

	err = GXNewJob(&docJob);
	
	
// If there are no errors, create a handle the size of our document structure and store
// the print job and page shape in it.  Store the handle in the window's refCon field so
// that we can get at it.  (Note that the utility routines "GetDocJob" and "GetDocShape"
// can be used to do this easily.

	if ( err == noErr )
	{
		windDoc = (TH_Doc) NewHandleClear(sizeof(T_Doc));

		if ( windDoc == NULL )
			err = MemError();
		else
		{
			(*windDoc)->docJob = docJob;
			(*windDoc)->docPage = docPage;
			SetWRefCon(wind, (long) windDoc);
		}

// Now install our application override for PrintingEvent so that we can
// support the new movable-modal printing dialog boxes.

		GXInstallApplicationOverride(docJob, gxPrintingEventMsg, gOurPrintingOverrideUPP);

	}

	return err;
}


/**\
|**| ---------------------------------------------------------------------
|**| CreateSampleImage()
|**| This function creates primitive shapes and adds them to the window's page shape.
|**| ---------------------------------------------------------------------
\**/
void CreateSampleImage (WindowPtr wind)
{
	Rect	ourWindowRect = wind->portRect; // the rectangle for this window
										    // in QuickDraw coordinates
	gxShape layout;						// the layout shape we build
	gxShape fromLayout;					// glyphs extracted from layout
	gxShape dumbText;					// a regular text shape
	gxShape fromDumbText;				// layout made from dumbText
	gxShape glyphs;						// just some glyphs
	gxShape fromGlyphs;					// layout made from glyph shape
	gxShape text;						// another text shape
	gxShape fromText;					// a layout made from a text shape
	gxShape thePage;					// this window's document shape
	gxRunControls runControls;			// run controls for the layout shape
	gxLayoutOptions layoutOptions;		// options for the layout shape
	gxStyle timesStyle;					// a style record for a Times-based style
	gxStyle helveticaStyle;				// ditto for Helvetica
	gxStyle baghdadStyle; 				// ditto for Baghdad
	gxStyle courierStyle; 				// ditto for Courier
	gxStyle textStyles[3];				// an array of text style for our text runs
	
// These are the four text runs for our shapes, in pieces because they're
// in different languages

	char *text0 = "Just a bunch of text!";
	
	char *text1 = "Aetna ";
	
// The following is "Arabic Macintosh" in Arabic: 
// meem, alif, kaf,  noon, tah,  wau,  shin, <sp>, alif, lam,  ein,  reh,  beh,  yeh
// (This assumes the standard Arabic character set for Macintosh, by the way)
	
	static char text2[] = {0xE5, 0xC7, 0xE3, 0xE6, 0xCA, 0xE8, 0xD4, 
						   0x20, 0xC7, 0xE4, 0xD9, 0xD1, 0xC8, 0xEA, 0};
	
	char *text3 = " Office AWAY.";

	char *textRuns[3];					// an array of pointers to text runs

	short textLengths[3];				// an array of the lengths of each run
	short totalLength;					// the total length of the layout's text
	short level0 = 0;					// This is so we can use its address
	long  advances = 0x80000000;		// advance bits for glyph shapes
	gxPoint textPositions[21] = {{ff(50), ff(50)}, 
								 {0, 0}, {0, 0}, {0, 0}, {0, 0},
								 {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
								 {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
								 {0, 0}, {0, 0}};	// positions for glyph shapes
								 
	gxPoint posn;						// the position of the layout shape


// make default gxLayoutOptions and gxRunControls structures

	InitializeLayoutOptions (&layoutOptions);
	InitializeRunControls (&runControls);
	
	
// Position the layout half way down the left edge of the window. Set 
// the layout's width to the window's width and set the flushness to 1/2, this
// will cause the layout to center in the window.  Note that ourWindowRect is
// not in fixed coordinates, so we have to make it fixed to use it.

	layoutOptions.width = ff(ourWindowRect.right - ourWindowRect.left);

	posn.x = 0;
	posn.y = ff((ourWindowRect.bottom - ourWindowRect.top) / 2);
	

// Create the styles we'll use.  All styles are 38-point.
// NewLayoutStyle is a library routine found in layout library.c.
	
	helveticaStyle = NewLayoutStyle(
		(char *) "\pHelvetica",				// the gxFontName for the style
		ff(38),								// the text size in fixed point
		0,									// gxTextAttributes
		&runControls,						// run controls (our default ones)
		nil,								// run features (none for this style)
		0,									// count of run features
		nil);								// style run overrides (none right now)

	baghdadStyle = NewLayoutStyle(
		(char *) "\pBaghdad Plain",			// the gxFontName for the style
		ff(38),								// the text size in fixed point
		0,									// gxTextAttributes
		&runControls,						// run controls (our default ones)
		nil,								// run features (none for this style)
		0,									// count of run features
		nil);								// style run overrides (none right now)

	timesStyle = NewLayoutStyle(
		(char *) "\pTimes Roman",			// the gxFontName for the style
		ff(38),								// the text size in fixed point
		0,									// gxTextAttributes
		&runControls,						// run controls (our default ones)
		nil,								// run features (none for this style)
		0,									// count of run features
		nil);								// style run overrides (none right now)

	courierStyle = NewLayoutStyle(
		(char *) "\pCourier",				// the gxFontName for the style
		ff(38),								// the text size in fixed point
		0,									// gxTextAttributes
		&runControls,						// run controls (our default ones)
		nil,								// run features (none for this style)
		0,									// count of run features
		nil);								// style run overrides (none right now)


// Get the shape for this window's page so we can add to it

	thePage = GetDocShape(wind);
	
// Initialize totalLength and create our first shape -- a plain text shape

	totalLength = MyStrLength(text0);
	text = GXNewText(
		totalLength,						// the length of the text
		(unsigned char *) text0,			// the text
		textPositions);						// positions of the glyphs
		
	GXSetShapeStyle(text, courierStyle);
	AddToShape(thePage, text);
	
// Make a copy of the text shape, move it down 50 points and change it to a layout

	fromText = GXCopyToShape(nil, text);
	textPositions[0].y += ff(50);			// move it down 50 points
	GXSetText(
		fromText,							// the shape to set the text in
		0,									// copy all chars in existing shape
		nil,								// don't change the text
		textPositions);						// the new place to put the shape
	GXSetShapeType(fromText, gxLayoutType);	// make it a layout shape
	AddToShape(thePage, fromText);			// and add it to the page
	
// Initialize the textLengths array to have about 1/3 in each of three styles

	textLengths[0] = totalLength/3;
	textLengths[1] = totalLength/3;
	textLengths[2] = totalLength - (textLengths[0] + textLengths[1]);
	
	textStyles[0] = courierStyle;
	textStyles[1] = helveticaStyle;
	textStyles[2] = timesStyle;

// Move down 50 more points and make a new glyph shape from our data

	textPositions[0].y += ff(50);			// move down 50 points
	
	glyphs = GXNewGlyphs(
		totalLength,						// the length of the text
		(void *) text0,						// the text to make glyphs from
		textPositions,						// where to put the glyphs
		&advances,							// the array of advance bits
		nil,								// no tangents
		textLengths,						// how many bytes in each style run
		textStyles);						// the array of text styles
		
	AddToShape(thePage, glyphs);
	
// Copy this to another shape and make it a layout shape like before

	fromGlyphs = GXCopyToShape(nil, glyphs);	// copy the shape
	textPositions[0].y += ff(50);				// move down 50 more points
	GXSetGlyphs(
		fromGlyphs,								// the shape to set glyphs in
		0,										// don't change char count
		nil,									// don't change text
		textPositions,							// new positions array
		nil,									// don't change advance bits
		nil,									// don't change tangents
		nil,									// don't change style run lengths
		nil);									// don't change styles
	GXSetShapeType(fromGlyphs, gxLayoutType);	// make it a layout shape
	AddToShape(thePage, fromGlyphs);			// and add it to the page
	
// Now it's time to really build a layout shape from scratch.  First, initialize
// the textRuns array to point to the text/

	textRuns[0] = text1;
	textRuns[1] = text2;
	textRuns[2] = text3;
	
// Now the textLengths array

	textLengths[0] = MyStrLength(text1);
	textLengths[1] = MyStrLength(text2);
	textLengths[2] = MyStrLength(text3);
	
	totalLength = textLengths[0] + textLengths[1] + textLengths[2];

// Initialize the textStyles array so that each of the runs has the right style

	textStyles[0] = helveticaStyle;
	textStyles[1] = baghdadStyle;
	textStyles[2] = timesStyle;
	
	textPositions[0].y += ff(50);		// move down 50 more points

// Setup complete!  Build the layout.

	layout = GXNewLayout(
		3,								// count of text runs
		textLengths,					// array of lengths of each run
		(void *) textRuns,				// array of pointers to the text
		3,								// count of style runs
		textLengths,					// array of the byte lengths of each run
		textStyles,						// array of the styles
		1,								// number of levels in the layout
		&totalLength,					// array of the lengths of each level
		&level0,						// array of the levels
		&layoutOptions,					// options (default)
		textPositions);					// position of the layout shape

	AddToShape(thePage, layout);

// Now copy the layout and turn it into a glyph shape.

	fromLayout = GXCopyToShape(nil, layout);
	textPositions[0].y += ff(50);		// move down 50 more points
	GXSetLayout(
		fromLayout,						// the layout shape to modify
		0,								// don't change text runs
		nil,
		nil,
		0,								// don't change style runs
		nil,
		nil,
		0,								// don't change levels
		nil,	
		nil,
		nil,							// don't change layout options
		textPositions);					// move to new position
	GXSetShapeType(fromLayout, gxGlyphType); // change to glyph shape
	AddToShape(thePage, fromLayout);
	
// Make a new dumb text shape in Baghdad to show how poorly these shapes treat
// this kind of stuff.

	textPositions[0].y += ff(50);		// down 50 more points
	dumbText = GXNewText(
		textLengths[1],					// the length of the Baghdad run
		(unsigned char *)text2,			// the Arabic text
		textPositions);					// the position of the shape
	GXSetShapeStyle(dumbText, baghdadStyle); // set the style to Baghdad
	AddToShape(thePage, dumbText);
	
// Now convert it to a layout shape and watch the difference

	textPositions[0].y += ff(50);		// down 50 _more_ points
	fromDumbText = GXCopyToShape(nil, dumbText);	// make a copy of the shape
	GXSetShapeType(fromDumbText, gxLayoutType);		// change it to a layout
	GXSetLayout(
		fromDumbText,					// the layout shape to modify
		0,								// don't change text runs
		nil,
		nil,
		0,								// don't change style runs
		nil,
		nil,
		0,								// don't change levels
		nil,	
		nil,
		nil,							// don't change layout options
		textPositions);					// move to new position
	AddToShape(thePage, fromDumbText);
	
// Now that all shapes are added to the picture, dispose of them

	GXDisposeShape(text);
	GXDisposeShape(fromText);
	GXDisposeShape(glyphs);
	GXDisposeShape(fromGlyphs);
	GXDisposeShape(layout);
	GXDisposeShape(fromLayout);
	GXDisposeShape(dumbText);
	GXDisposeShape(fromDumbText);
	GXDisposeStyle (courierStyle);
	GXDisposeStyle (baghdadStyle);
	GXDisposeStyle (helveticaStyle);
	GXDisposeStyle (timesStyle);
	

// Invalidate the window's portRect so that everything gets updated.
	
	SetPort(wind);
	InvalRect(&ourWindowRect);
}

